#include "CGame.h"

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image/stb_image.h"
#include "Model.h"
#include "Mesh.h"
#include "glm/mat4x4.hpp"
#include "glm/vec3.hpp"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp"
#include "glm/gtc/type_ptr.hpp"
#include <chrono>

void CGame::Initialize(unsigned int screenWidth, unsigned int screenHeight)
{
    m_Width = screenWidth;
    m_Height = screenHeight;

    SetupGPUProgram();
    SetupModel();
    SetupFullscreenQuad();

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_STENCIL_TEST);
    glEnable(GL_MULTISAMPLE);

    GLuint depthStencilId;

    glGenFramebuffers(1, &m_IntermediateFBOId);
    glBindFramebuffer(GL_FRAMEBUFFER, m_IntermediateFBOId);

    glGenTextures(1, &m_ColorTextureId);
    glBindTexture(GL_TEXTURE_2D_MULTISAMPLE, m_ColorTextureId);
    glTexImage2DMultisample(GL_TEXTURE_2D_MULTISAMPLE, 4, GL_RGB8, m_Width, m_Height, GL_TRUE);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D_MULTISAMPLE, m_ColorTextureId, 0);

    glGenRenderbuffers(1, &depthStencilId);
    glBindRenderbuffer(GL_RENDERBUFFER, depthStencilId);
    glRenderbufferStorageMultisample(GL_RENDERBUFFER, 4, GL_DEPTH24_STENCIL8, screenWidth, screenHeight);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_RENDERBUFFER, depthStencilId);

    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) == GL_FRAMEBUFFER_COMPLETE)
    {
        // Great!
    }

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void CGame::Shutdown()
{
    
}

void CGame::OnResize(int width, int height)
{
    m_Width = width;
    m_Height = height;

    glViewport(0, 0, m_Width, m_Height);
    glScissor(0, 0, m_Width, m_Height);
}

void CGame::Render()
{
    static auto startTime = std::chrono::high_resolution_clock::now();

    auto currentTime = std::chrono::high_resolution_clock::now();
    float time = std::chrono::duration<float, std::chrono::seconds::period>(currentTime - startTime).count();
	time = 0.0f;

    float camYOffset = -0.5f;
    float camZOffset = 5.0f;

    glEnable(GL_MULTISAMPLE);
    glBindFramebuffer(GL_FRAMEBUFFER, m_IntermediateFBOId);

    glClearColor(0.0f, 1.0f, 0.0f, 1.0f);
    glClearDepth(1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Draw the Filled Man
    glm::mat4 model = glm::rotate(glm::mat4(1.0f), time * glm::radians(90.0f), glm::vec3(0.0f, 1.0f, 0.0f));
    glm::mat4 view;
    view = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -2.0f + camYOffset, -10.0f + camZOffset));
    glm::mat4 projection;
    projection = glm::perspective(glm::radians(60.0f), m_Width / static_cast<float>(m_Height), 0.03f, 100.0f);
    glm::mat4 mvp = projection * view * model;

    m_pShaderPass0->use();
    m_pShaderPass0->setMatrix4x4("modelMatrix", glm::value_ptr(model));
    m_pShaderPass0->setMatrix4x4("viewMatrix", glm::value_ptr(view));
    m_pShaderPass0->setMatrix4x4("projectionMatrix", glm::value_ptr(projection));
    m_pShaderPass0->setMatrix4x4("mvpMatrix", glm::value_ptr(mvp));
    m_pShaderPass0->setFloat("time", time);
    glDepthFunc(GL_LEQUAL);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    m_pManModel->meshes[0].Draw(*m_pShaderPass0);

    // Blit intermediate FBO to backbuffer FBO
    glBindFramebuffer(GL_FRAMEBUFFER, 0);

    m_pShaderBlitPass->use();
    glActiveTexture(0);
    glBindTexture(GL_TEXTURE_2D_MULTISAMPLE, m_ColorTextureId);

    glDepthFunc(GL_ALWAYS);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    m_pFullscreenQuad->Draw(*m_pShaderBlitPass);
}

void CGame::SetupModel()
{
    m_pManModel = new Model();
    m_pManModel->Load("man.obj");

    m_pStencilModel = new Model();
    m_pStencilModel->Load("quad.obj");
}

void CGame::SetupGPUProgram()
{
    m_pShaderPass0 = new Shader("vertexPass0.glsl", "fragmentPass0.glsl");
    m_pShaderBlitPass = new Shader("vertexBlit.glsl", "fragmentBlit.glsl");
}

void CGame::SetupFullscreenQuad()
{
    vector<Vertex> vertices;

    float positions[4][3] =
    {
        { -1, -1, 0.3f },
        { -1, 1, 0.3f },
        { 1, 1, 0.3f },
        { 1, -1, 0.3f }
    };

    float uvs[4][2] =
    {
        { 0,0 },
        { 0,1 },
        { 1,1 },
        { 1,0 }
    };

    float indicesVals[6] =
    {
        0,1,2,
        0,2,3
    };

    for (int i = 0; i < 4; ++i)
    {
        Vertex vertex;
        vertex.Position.x = positions[i][0];
        vertex.Position.y = positions[i][1];
        vertex.Position.z = positions[i][2];

        vertex.TexCoords.x = uvs[i][0];
        vertex.TexCoords.y = uvs[i][1];

        vertices.push_back(vertex);
    }

    vector<unsigned int> indices(indicesVals, indicesVals + 6);

    m_pFullscreenQuad = new Mesh(vertices, indices);
}